#!/usr/bin/env -S python3 -u
#-*- coding:utf-8 -*-
#
########################################################################
#
# Read fastNLO grids and print cross section & uncertainty
#
# Created by K. Rabbertz, 22.04.2024
#
########################################################################
#
import argparse
import glob
import os
import re
import string
import sys
import timeit
# numpy
import numpy as np
# fastNLO for direct evaluation of interpolation grids
# ATTENTION: fastNLO python extension is required for Python 3!
import fastnlo
from fastnlo import fastNLOLHAPDF
from fastnlo import SetGlobalVerbosity

# Action class to allow comma-separated (or empty) list in options
class SplitArgs(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            setattr(namespace, self.dest, values[0].split(','))
        else:
            setattr(namespace, self.dest, [''])


# Some global definitions
#_fntrans = str.maketrans({'[': '', ']': '', '(': '', ')': '', ',': ''}) # Filename translation table
#_sntrans = str.maketrans({'[': '', ']': '', '(': '', ')': '', ',': '', '/': '÷'}) # Scalename translation table
#_formats = {'eps': 0, 'pdf': 1, 'png': 2, 'svg': 3}
_text_to_order = {'LO': 0, 'NLO': 1, 'NNLO': 2}
_order_to_text = {0: 'LO', 1: 'NLO', 2: 'NNLO'}
#_order_color = {'LO': 'g', 'NLO': 'b', 'NNLO': 'r'}
#_order_symbol = {'LO': '^', 'NLO': 's', 'NNLO': 'o'}
#_colors = ['tab:orange', 'tab:green', 'tab:purple', 'tab:blue', 'tab:brown']
#_symbols = ['s', 'X', 'o', '^', 'v']
#_hatches = ['', '//', '\\', '|', '-']
#_scale_to_text = {0: 'kScale1', 1: 'kScale2', 2: 'kQuadraticSum', 3: 'kQuadraticMean', 4: 'kQuadraticSumOver4',
#                  5: 'kLinearMean', 6: 'kLinearSum', 7: 'kScaleMax', 8: 'kScaleMin', 9: 'kProd',
#                  10: 'kS2plusS1half', 11: 'kPow4Sum', 12: 'kWgtAvg', 13: 'kS2plusS1fourth', 14: 'kExpProd2', 15: 'kExtern'}
#_pdfbasenames = ['ABM11', 'ABMP15', 'ABMP16', 'CJ12', 'CJ15', 'CT10', 'CT14', 'HERAPDF20', 'JR14',
#                 'MMHT2014', 'MSTW2008', 'NNPDF23', 'NNPDF30', 'NNPDF31']
#_debug = False
# TODO: How can this be done using fastNLO directly?
_verb_to_enum = {'SILENT': 1000, 'ERROR': 2, 'WARNING': 1, 'INFO': 0, 'MANUAL': -1000, 'DEBUG': -1000}


########################################################################

def main():
    # Start timer
    # just for measuring wall clock time - not necessary
    start_time = timeit.default_timer()
    # Define arguments & options
    parser = argparse.ArgumentParser(epilog='', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # Positional arguments
    parser.add_argument('table', type=str, nargs='+',
                        help='Filename glob of fastNLO tables to be evaluated. This must be specified!')
    # Optional arguments
    parser.add_argument('-m', '--member', default=0, type=int,
                        help='Member of PDFset, default is 0.')
    parser.add_argument('-o', '--order', required=False, nargs=1, type=str, action=SplitArgs,
                        help='Comma-separated list of orders to show: LO, NLO, and/or NNLO. If nothing is chosen, show all orders available in table.')
    parser.add_argument('-p', '--pdfset', default='CT14nlo', type=str,
                        help='PDFset to evaluate fastNLO table.')
    parser.add_argument('-s', '--scale', default=0, required=False, nargs='?', type=int,
                        choices=list(range(16)), metavar='[0-15]',
                        help='For flexible-scale tables define central scale choice for MuR and MuF by selection enum fastNLO::ScaleFunctionalForm ("0"=kScale1, "1"=kScale2, "2"=kQuadraticSum), ...')
    parser.add_argument('-u', '--uncertainty', default=None, required=False, type=str,
                        help='Type of uncertainty to be derived: 2P, 6P, L6, ST.')
    parser.add_argument('-v', '--verbosity', default='WARNING', type=str,
                        help="Set fastNLO output verbosity to one of SILENT, ERROR, WARNING, INFO, MANUAL, DEBUG")

    # Parse arguments
    args = vars(parser.parse_args())

    # Verbosity
    verb = args['verbosity']
    eVerb = _verb_to_enum[verb]
    #    fVerbosity = toVerbosity()[VerbosityLevel]
    SetGlobalVerbosity(eVerb)

    # Print header
    if eVerb < 1:
        print(" ##################################################################################")
        print(" # [fnlo-py-print] Program to read fastNLO tables and write out")
        print(" # [fnlo-py-print] QCD cross sections with uncertainties")
        print(" #---------------------------------------------------------------------------------")
        print(" # [fnlo-py-print] For an explanation of command line arguments type:")
        print(" # [fnlo-py-print] ./fnlo-py-print.py -h")
        print(" ##################################################################################")
        print("")

    # List of table names
    files = args['table']
    if eVerb < 1000:
        print(" ##################################################################################")
        print(" # [fnlo-py-print] Iterating over table list:")
    for file in files:
        if eVerb < 1000:
            print(' # [fnlo-py-print]   ', file)

    # PDF set name
    pdfset = os.path.basename(args['pdfset'])
    if eVerb < 1000:
        print(' # [fnlo-py-print] Using PDF set:', pdfset)

    # Orders to be shown
    iorders = []
    iordmin = _text_to_order['LO']
    iordmax = _text_to_order['NNLO']
    if args['order'] is None:
        if eVerb < 1000:
            print(' # [fnlo-py-print] Evaluate tables up to highest available order.')
    else:
        for ord in args['order']:
            if ord in _text_to_order:
                iorders.append(_text_to_order[ord])
            else:
                print(' # [fnlo-py-print] Illegal order specified, aborted!')
                print(' # [fnlo-py-print] Order list:', args['order'])
                exit(1)
        iordmin = min(iorders)
        iordmax = max(iorders)
        if eVerb < 1000:
            print(' # [fnlo-py-print] Evaluate tables up to order(s):', args['order'])
            print(" ##################################################################################")

    # Type of uncertainty (None, scale variation [2P,6P], PDF [L6], statistical [ST])
    unc_type  = None
    unc_style = fastnlo.kScaleNone
    unc_label = ''
    if args['uncertainty']:
        unc_type = args['uncertainty']
        if unc_type == '2P':
            unc_style = fastnlo.kSymmetricTwoPoint
            unc_label = 'Scale uncertainty (2P)'
        elif unc_type == '6P':
            unc_style = fastnlo.kAsymmetricSixPoint
            unc_label = 'Scale uncertainty (6P)'
        elif unc_type == 'L6':
            unc_style = fastnlo.kLHAPDF6
            unc_label = 'PDF uncertainty (LHAPDF)'
        elif unc_type == 'ST':
            unc_style = fastnlo.kStatInt
            unc_label = 'numerical uncertainty (theory)'
        else:
            print(' # [fnlo-py-print] Illegal uncertainty specified, aborted!')
            print(' # [fnlo-py-print] Uncertainty:', args['uncertainty'])
            exit(11)

    # Scale choice
    scale_choice = args['scale']

    # Loop over table list
    for table in files:
        # Table name
        tablepath = os.path.split(table)[0]
        if not tablepath:
            tablepath = '.'
        tablename = os.path.split(table)[1]
        if tablename.endswith('.tab.gz'):
            tablename = tablename.replace('.tab.gz', '', 1)
        elif tablename.endswith('.tab'):
            tablename = tablename.replace('.tab', '', 1)
        else:
            print(' # [fnlo-py-print] Error! Wrong extension for table: ', table)
            exit(1)
        if eVerb < 1000:
            print("")
            print(" ##################################################################################")
            print(' # [fnlo-py-print] Evaluating table: ', table)
            print(" ##################################################################################")
            print("")

        ###################### Start EVALUATION with fastNLO library ###################################################
        # Take the general information (bin_bounds, labels, order_existence, etc.) from given pdfset.
        fnlo = fastnlo.fastNLOLHAPDF(table, args['pdfset'], args['member'])

        # Get labeling for the x-axis
        # Dimensionality of the table:
        ndim = fnlo.GetNumDiffBin()
        if eVerb < 0:
            print('')
            print(' # [fnlo-py-print] Table Dimensions: ', ndim)

        # Labels of all the dimensions:
        labels = fnlo.GetDimLabels()
        if eVerb < 0:
            print(' # [fnlo-py-print] Labels:', labels)

        # x label of first dimension from table:
        xlabel = fnlo.GetDimLabel(0)
        if eVerb < 0:
            print(' # [fnlo-py-print] x-label:', xlabel)

        # Generic y label
        ylabel = '$\sigma \pm \Delta\sigma(\mu_R,\mu_F)$'

        # Creating x-axis
        bin_bounds = np.array(fnlo.GetObsBinsBounds(0))
        #        bin_bounds = np.array(fnlo.GetDim1BinBounds(0)) Mods needed for 2D tables ...
        if eVerb < 0:
            print(' # [fnlo-py-print] bin_bounds.T: \n', bin_bounds.T, '\n')
            print(' # [fnlo-py-print] bin_bounds.flatten()', bin_bounds.flatten(), '\n')

        x_axis = (bin_bounds.T[0]+bin_bounds.T[1]) / 2.  # this is a list of bin centers
        xmin = 0.95*min(bin_bounds.ravel())
        xmax = 1.05*max(bin_bounds.ravel())
        if eVerb < 0:
            print(' # [fnlo-py-print] xmin=%s, xmax=%s. \n' % (xmin, xmax))

        # Preparing x-errors (via bin_bounds) --> x_errors[0, :] are initially negative (positive via -1*), x_errors[1, :] positive
        x_errors = np.array(
            [-1*(bin_bounds.T[0]-x_axis), bin_bounds.T[1]-x_axis])
        if eVerb < 0:
            print(' # [fnlo-py-print] \n x_errors: ', x_errors, '\n')

        # Check existence of orders in table
        lflex = fnlo.GetIsFlexibleScaleTable()
        scale_name = 'scale1'
        o_existence = [False, False, False]
        cnt_order = -1
        max_order = 0
        for i in range(3):
            o_existence[i] = fnlo.SetContributionON(
                fastnlo.kFixedOrder, i, True)
            if o_existence[i]:
                max_order = i
                if not lflex:
                    if scale_choice != 0:
                        print(' # [fnlo-py-print] Invalid choice of scale = ', scale_choice, ' Aborted!')
                        print(' # [fnlo-py-print] For fixed-scale tables only the default=0 is allowed.')
                        exit(1)
                    else:
                        scale_name = fnlo.GetScaleDescription(i, 0)
                else:
                    if scale_choice < 2:
                        scale_name = fnlo.GetScaleDescription(i, scale_choice)
                    else:
                        scl0 = fnlo.GetScaleDescription(i, 0)
                        scl1 = fnlo.GetScaleDescription(i, 1)
                        scale_name = _scale_to_text[scale_choice] + \
                            '_'+scl0+'_'+scl1
                if cnt_order == i-1:
                    cnt_order += 1
        if eVerb < 0:
            print(' # [fnlo-py-print] Table has continuous orders up to',
                  cnt_order, 'and a maximal order of', max_order)

        # If previously undefined set iordmax to maximum found in table
        if args['order'] is None:
            iordmax = max_order

        if iordmax > cnt_order:
            print(' # [fnlo-py-print] Invalid choice of orders. Aborted!')
            print(' # [fnlo-py-print] Highest order requested is',
                  _order_to_text[iordmax], 'but continuous orders are available only up to', _order_to_text[cnt_order])
            exit(1)

        order_list = []
        if args['order'] is None:
            for iord in range(cnt_order+1):
                order_list.append(_order_to_text[iord])
        else:
            order_list = args['order']

        if eVerb < 1:
            print(' # [fnlo-py-print] List of requested orders:', order_list)

        # For flexible-scale tables set scale to user choice (default is 0)

        if lflex:
            if eVerb < 0:
                print(' # [fnlo-py-print] Setting requested scale choice for flexible-scale table:', scale_choice)
            fnlo.SetMuRFunctionalForm(scale_choice)
            fnlo.SetMuFFunctionalForm(scale_choice)
        else:
            if scale_choice == 0:
                if eVerb < 0:
                    print(' # [fnlo-py-print] Evaluating fixed-scale table. Scale choice must be', scale_choice)
            else:
                print(' # [fnlo-py-print] No scale choice possible for fixed-scale table. Aborted!')
                print(' # [fnlo-py-print] scale_choice = ', scale_choice)
                exit(1)

        # Now evaluate fastNLO table having a look at the requested uncertainty
        xs_list = []  # will contain total cross section for selected orders out of LO, NLO, NNLO
        # list for relative scale uncertainties (low, high) for selected orders
        rel_unc_list = []
        for n in order_list:
            for j in range(0, max_order+1):
                if j <= _text_to_order[n]:
                    fnlo.SetContributionON(fastnlo.kFixedOrder, j, True)
                else:
                    fnlo.SetContributionON(fastnlo.kFixedOrder, j, False)
            if eVerb < 0:
                print(' # [fnlo-py-print] \n')
                print(' # [fnlo-py-print] Calculate XS for order: %s' % n, '\n')
                print(' # [fnlo-py-print] ----  ----  ----  ----  ----  ----  ----  ----')

            fnlo.CalcCrossSection()
            xs_list.append(fnlo.GetCrossSection())

            ### Get requested uncertainty ###
            # RELATIVE uncertainty for chosen uncertainty type and style
            # Calculate this for all accessible orders just in case
            if eVerb < 0:
                print(' # [fnlo-py-print] Used scale factor MuF: ', fnlo.GetScaleFactorMuF())
                print(' # [fnlo-py-print] Used scale factor MuR: ', fnlo.GetScaleFactorMuR(), '\n')
                print(' # [fnlo-py-print] Calculate uncertainty for this central scale.\n')
            elif eVerb < 1000:
                print(" ##################################################################################")
                print("")
            # Start of fastNLO print out
            lnorm  = False
            iprint = 1
            if unc_type == '2P' or unc_type == '6P':
                rel_unc_item = np.array(fnlo.GetXsUncertaintyVec(unc_style,lnorm,iprint))
            elif unc_type == 'L6':
                rel_unc_item = np.array(fnlo.GetXsUncertaintyVec(unc_style,lnorm,iprint))
            elif unc_type == 'ST':
                rel_unc_item = np.array(fnlo.GetXsUncertaintyVec(unc_style,lnorm,iprint))
            else:
                rel_unc_item = np.array(fnlo.GetXsUncertaintyVec(fastnlo.kScaleNone,lnorm,iprint))

            rel_unc_list.append(rel_unc_item)

        xs_all = np.array(xs_list)
        rel_unc = np.array(rel_unc_list)
        ##########
        # structure of rel_unc:
        # rel_unc[0,:,:] means LO, rel_unc[1,:,:] means NLO, and rel_unc[2,:,:] in NNLO
        # rel_unc[0,0,:] means xs in LO
        # rel_unc[0,1,:] means rel. uncertainty upwards (in LO)
        # rel_unc[0,2,:] means rel. uncertainty downwards (in LO)
        #########

        stop_time = timeit.default_timer()
        timediff = stop_time-start_time
        if eVerb < 1:
            print(' # [fnlo-py-print] Elapsed time: %s sec = %s min' % (timediff, round(timediff/60, 2)))


if __name__ == '__main__':
    main()
